/*============================================================================
KEYBOARD ROUTINES

EXPORTS:
void keyboard_irq(void);
void init_keyboard(void);
============================================================================*/
#include <conio.h> /* KEY_nnn */
#include <x86.h> /* outportb(), inportb() */
#include "_krnl.h" /* MAX_VC */
#include <time.h>


/* IMPORTS
from VIDEO.C */
extern console_t _vc[];

void select_vc(unsigned which_vc);
void putch(unsigned c);
int search_comnd(char*); // searches inputted command string for valid command word


/* from MAIN.C */
void kprintf(const char *fmt, ...);

/* "raw" set 1 scancodes from PC keyboard.*/
#define	RAW1_LEFT_CTRL		0x1D
#define	RAW1_RIGHT_CTRL		0x1D	/* same as left */
#define	RAW1_LEFT_SHIFT		0x2A
#define	RAW1_RIGHT_SHIFT	0x36
#define	RAW1_LEFT_ALT		0x38
#define	RAW1_RIGHT_ALT		0x38	/* same as left */
#define	RAW1_CAPS_LOCK		0x3A
#define	RAW1_F1			0x3B
#define	RAW1_F2			0x3C
#define	RAW1_F3			0x3D
#define	RAW1_F4			0x3E
#define	RAW1_F5			0x3F
#define	RAW1_F6			0x40
#define	RAW1_F7			0x41
#define	RAW1_F8			0x42
#define	RAW1_F9			0x43
#define	RAW1_F10		0x44
#define	RAW1_NUM_LOCK		0x45
#define	RAW1_SCROLL_LOCK	0x46
#define	RAW1_F11		0x57
#define	RAW1_F12		0x58

#define	KBD_BUF_SIZE		64
/*****************************************************************************
*****************************************************************************/
void delay()
{
	long cnt1;
	long cnt2;

    cnt1=999999999;
	cnt2=100;

	while(cnt1!=0)
	{
		while(cnt2!=0)
		{
			cnt2--;
		}

		cnt1--;
	}

	return;
}

/*****************************************************************************
*****************************************************************************/
static void reboot(void)
{
	unsigned temp;

	disable();
/* flush the keyboard controller */
	do
	{
		temp = inportb(0x64);
		if((temp & 0x01) != 0)
		{
			(void)inportb(0x60);
			continue;
		}
	} while((temp & 0x02) != 0);
/* pulse the CPU reset line */
	outportb(0x64, 0xFE);
/* ...and if that didn't work, just halt */
	while(1)
		/* nothing */;
}
/*****************************************************************************
*****************************************************************************/
static void write_kbd(unsigned adr, unsigned data)
{
	unsigned long timeout;
	unsigned stat;

/* Linux code didn't have a timeout here... */
	for(timeout = 500000L; timeout != 0; timeout--)
	{
		stat = inportb(0x64);
/* loop until 8042 input buffer empty */
		if((stat & 0x02) == 0)
			break;
	}
	if(timeout != 0)
		outportb(adr, data);
}
/*****************************************************************************
*****************************************************************************/
static unsigned convert(unsigned key)
{
	
	static const unsigned char set1_map[] =
	{
/* 00 */0,	0x1B,	'1',	'2',	'3',	'4',	'5',	'6',
/* 08 */'7',	'8',	'9',	'0',	'-',	'=',	'\b',	'\t',
/* 10 */'q',	'w',	'e',	'r',	't',	'y',	'u',	'i',
/* 1Dh is left Ctrl */
/* 18 */'o',	'p',	'[',	']',	'\n',	0,	'a',	's',
/* 20 */'d',	'f',	'g',	'h',	'j',	'k',	'l',	';',
/* 2Ah is left Shift */
/* 28 */'\'',	'`',	0,	'\\',	'z',	'x',	'c',	'v',
/* 36h is right Shift */
/* 30 */'b',	'n',	'm',	',',	'.',	'/',	0,	0,
/* 38h is left Alt, 3Ah is Caps Lock */
/* 38 */0,	' ',	0,	KEY_F1,	KEY_F2,	KEY_F3,	KEY_F4,	KEY_F5,
/* 45h is Num Lock, 46h is Scroll Lock */
/* 40 */KEY_F6,	KEY_F7,	KEY_F8,	KEY_F9,	KEY_F10,0,	0,	KEY_HOME,
/* 48 */KEY_UP,	KEY_PGUP,'-',	KEY_LFT,'5',	KEY_RT,	'+',	KEY_END,
/* 50 */KEY_DN,	KEY_PGDN,KEY_INS,KEY_DEL,0,	0,	0,	KEY_F11,
/* 58 */KEY_F12
	};
	static unsigned short kbd_status, saw_break_code;
/**/
	unsigned short temp;

/* check for break key (i.e. a key is released) */
	if(key >= 0x80)
	{
		saw_break_code = 1;
		key &= 0x7F;
	}
/* the only break codes we're interested in are Shift, Ctrl, Alt */
	if(saw_break_code)
	{
		if(key == RAW1_LEFT_ALT || key == RAW1_RIGHT_ALT)
			kbd_status &= ~KBD_META_ALT;
		else if(key == RAW1_LEFT_CTRL || key == RAW1_RIGHT_CTRL)
			kbd_status &= ~KBD_META_CTRL;
		else if(key == RAW1_LEFT_SHIFT || key == RAW1_RIGHT_SHIFT)
			kbd_status &= ~KBD_META_SHIFT;
		saw_break_code = 0;
		return 0;
	}
/* it's a make key: check the "meta" keys, as above */
	if(key == RAW1_LEFT_ALT || key == RAW1_RIGHT_ALT)
	{
		kbd_status |= KBD_META_ALT;
		return 0;
	}
	if(key == RAW1_LEFT_CTRL || key == RAW1_RIGHT_CTRL)
	{
		kbd_status |= KBD_META_CTRL;
		return 0;
	}
	if(key == RAW1_LEFT_SHIFT || key == RAW1_RIGHT_SHIFT)
	{
		kbd_status |= KBD_META_SHIFT;
		return 0;
	}
/* Scroll Lock, Num Lock, and Caps Lock set the LEDs. These keys
have on-off (toggle or XOR) action, instead of momentary action */
	if(key == RAW1_SCROLL_LOCK)
	{
		kbd_status ^= KBD_META_SCRL;
		goto LEDS;
	}
	if(key == RAW1_NUM_LOCK)
	{
		kbd_status ^= KBD_META_NUM;
		goto LEDS;
	}
	if(key == RAW1_CAPS_LOCK)
	{
		kbd_status ^= KBD_META_CAPS;
LEDS:		write_kbd(0x60, 0xED);	/* "set LEDs" command */
		temp = 0;
		if(kbd_status & KBD_META_SCRL)
			temp |= 1;
		if(kbd_status & KBD_META_NUM)
			temp |= 2;
		if(kbd_status & KBD_META_CAPS)
			temp |= 4;
		write_kbd(0x60, temp);	/* bottom 3 bits set LEDs */
		return 0;
	}
/* ignore invalid scan codes */
	if(key >= sizeof(set1_map) / sizeof(set1_map[0]))
		return 0;
/* convert raw scancode in key to unshifted ASCII in temp */
	temp = set1_map[key];
/* defective keyboard? non-US keyboard? more than 104 keys? */
	if(temp == 0)
		return temp;
/* handle the three-finger salute */
	if((kbd_status & KBD_META_CTRL) && (kbd_status & KBD_META_ALT) && (temp == KEY_DEL))
	{
				kprintf("\x1B[2J""\n\n\n\n\n\n\n\n\n\n");
				kprintf("***  Good bye  *** \n");
				kprintf("***  Please Wait while AVOS Reboots ...");


                delay();

                reboot();
	}

	return temp;
}
/*****************************************************************************
*****************************************************************************/
void keyboard_irq(void)
{

	struct date d;
	struct time t;
	unsigned key, i;
	static unsigned vc;
	static char cmd_buff[5];
	static unsigned buff_ptr;
	static unsigned bad_comnd;
	static char *weekday[] = { "Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"};
	int  comnd;
	bad_comnd=0;

/* get scancode from port 0x60 */
	key = inportb(0x60);
/* if it's F1, F2 etc. switch to the appropriate virtual console */
	switch(key)
	{
	case RAW1_F1:
		i = 0;
		goto SWITCH_VC;
	case RAW1_F2:
		i = 1;
		goto SWITCH_VC;
	case RAW1_F3:
		i = 2;
		goto SWITCH_VC;
	case RAW1_F4:
		i = 3;
		goto SWITCH_VC;
	case RAW1_F5:
		i = 4;
		goto SWITCH_VC;
	case RAW1_F6:
		i = 5;
		goto SWITCH_VC;
	case RAW1_F7:
		i = 6;
		goto SWITCH_VC;
	case RAW1_F8:
		i = 7;
		goto SWITCH_VC;
	case RAW1_F9:
		i = 8;
		goto SWITCH_VC;
	case RAW1_F10:
		i = 9;
		goto SWITCH_VC;
	case RAW1_F11:
		i = 10;
		goto SWITCH_VC;
	case RAW1_F12:
		i = 11;
SWITCH_VC:
		select_vc(i);
		vc=i;
		break;
	default:
		i = convert(key);
		if((i==KEY_HOME) || (i==0x09) || (i==KEY_INS) || (i==KEY_LWIN) || (i==KEY_RWIN) || (i==KEY_MENU) || (i==KEY_PRNT) || (i==KEY_PAUSE)|| (i==KEY_DEL) || (i==KEY_UP) || (i==KEY_PGUP) || (i==KEY_RT) || (i==KEY_END) || (i==KEY_DN) || (i==KEY_PGDN))
		{
			break;
		}
		
		if(((i==0x08) || (i==KEY_LFT)) && (buff_ptr!=0) ) // for back space
		{
			if(buff_ptr<=4)
				cmd_buff[buff_ptr-1]='\0';
			buff_ptr--;
			putch(i);
			break;
		}

		if((i != 0) && (i!= '\n') && (i!= 0x08))
		{

			if(buff_ptr<=4)
			{
				cmd_buff[buff_ptr]=i;
				
			}
			else if(buff_ptr > 4)
				bad_comnd=1;

			buff_ptr++;
			putch(i);
			break;
		}



		if(i=='\n' && (buff_ptr>0))
		{
			if(bad_comnd)
			{
				kprintf("\nInvalid Command\n");
				bad_comnd=0;
			}

			comnd=search_comnd(cmd_buff);
			
			switch(comnd)
				{

					case 1:
							kprintf("\x1B[2J Advanced OS ver: 1.2.0 \nYou are at Virtual Console: %u\n",vc+1);
							
							break;
					case 2:
							kprintf("\nAdvanced OS version 1.2.0\n");
							break;
					case 3:
							kprintf("\x1B[2J""\n\n\n\n\n\n\n\n\n\n\n\n");
							kprintf("***  Good bye  *** \n");
							kprintf("***  Please Wait while AVOS Reboots ...");
                            delay();
							reboot();
							break;
					case 4:
							kprintf("\x1B[2J""\n\n AVOS Help:\n----------\n\n");
							kprintf("F1		Console #1\n");
							kprintf("F2		Console #2\n");
							kprintf("F3		Console #3\n");
							kprintf("F4		Console #4\n");
							kprintf("F5		Console #5\n");
							kprintf("F6		Console #6\n");
							kprintf("F7		Console #7\n");
							kprintf("F8		Console #8\n");
							kprintf("F9		Console #9\n");
							kprintf("F10		Console #10\n");
							kprintf("F11		Console #11\n");
							kprintf("F12		Console #12\n\n");
							kprintf("cls	Clears Current Console Screen\n");
							kprintf("abt	About AVOS, technical information\n");
							kprintf("dat	Prints current Date\n");
							kprintf("tme	Prints current Time\n");
							kprintf("ver	Prints AVOS Version\n");
							kprintf("reb	Reboots Computer\n");
							kprintf("hlp	Displays this help screen\n");
							break;
					
					case 5: // for date display
					
							kprintf("\nCurrent date: ");
							getdate(&d);
							kprintf("%s, %1d%1d/%01d%1d/'%1d%1d\n", weekday[d.dayofweek-1], d.day>>4,d.day & 0x0F, d.month>>4,d.month & 0x0F, d.year>>4, d.year & 0x0F);
							break;

					case 6: // Time Display ...
							
        					kprintf("\nCurrent time: ");
							gettime(&t);
							kprintf("%1d%1d:%1d%1d:%1d%1d HRS\n", t.hour>>4, t.hour & 0x0F, t.min>>4,t.min & 0x0F, t.sec>>4, t.sec & 0x0F);
							break;

					case 7: // About AVOS, what it suppoorts, what not ....
							kprintf("\x1B[2J""\n AVOS:Advanced OS version 1.2.0\n");
							kprintf("Developed as the final semester project of Bachelor of Technology \n(Under Grad. Coursework ");
							kprintf ("with Computer Engg. as major).\n");
							kprintf("\nFeatures:\n------------------\n");
							kprintf("A. 8 to 12 virtual consoles (depending upon the framebuffer size) where seperate programs can be executed as multitasking applications --\n");
							kprintf("B. Only Character user interface (text based) console for accepting some basic  commands.\n");
							kprintf("C. Supports keyboard input of some basic commands.\n");
							kprintf("D. A rudimentary form of floppy driver has been written to detect and activate  the installed floppy drives.\n");
							kprintf("E. Following commands are supported: changing virtual consoles, clearing screen, display time, display date, rebooting machine, display help.\n");
							kprintf("\nDoes NOT support:\n------------------\n");
							kprintf("Any filesystem, FAT, FAT32, NTFS, ext2fs, befs, ext3fs etc.\n");
							kprintf("Any file related operations, creating, deleting, renaming, editing or copying,  moving files.\n");
							kprintf("Mouse support, CD Rom support, and other hardwares.\n");
							kprintf("===========================================================\n");
							kprintf("(C) Anirban Sinha, Satyajit Chakrabarti, 2003; All Rights Reserved.");

							break;
							
					case -1:
							kprintf("\nInvalid Command\n");
				}	

							
			putch(i);
			kprintf("$>"); // prompt
			buff_ptr=0;
			cmd_buff[0]='\0';
			cmd_buff[1]='\0';
			cmd_buff[2]='\0';
			cmd_buff[3]='\0';
			cmd_buff[4]='\0';
			break;
		}
		else if((i=='\n') && (buff_ptr==0)) // single enter key at prompt
		{
			putch(i);
			kprintf("$>");
			break;
		}
		
	}
/* reset hardware interrupt at 8259 chip */
	outportb(0x20, 0x20);
}
/*****************************************************************************
*****************************************************************************/
int search_comnd(char *comnd) // searches inputted command string for valid command word
{
	if(*(comnd+3)!='\0') // max length of commands is 3 characters
		return -1;

	switch(*(comnd+0))
	{
		case 'A': // About AVOS ....
		case 'a':
					switch(*(comnd+1))
					{
						case 'b':
						case 'B':
									switch (*(comnd+2))
									{
										case 't':
										case 'T':
													return 7; // About AVOS .....
													break;
										default:
													return -1;
									}
									break;
						default:
									return -1;
					}

					break;
		case 'T':
		case 't':
					switch(*(comnd+1))
					{
						case 'm':
						case 'M':
									switch (*(comnd+2))
									{
										case 'e':
										case 'E':
													return 6;
													break;
										default:
													return -1;
									}
									break;
						default:
									return -1;
					}

					break;

		case 'c':
		case 'C':
					switch(*(comnd+1))
					{
						case 'l':
						case 'L':
									switch (*(comnd+2))
									{
										case 's':
										case 'S':
													return 1;
													break;
										default:	return -1;
									}
						break;
						default: return -1;
					}
					break;

		case 'v':
		case 'V':
					switch(*(comnd+1))
					{
						case 'e':
						case 'E':
									switch (*(comnd+2))
									{
										case 'r':
										case 'R':
													return 2;
													break;
										default:	return -1;
									}
						break;
						default: return -1;
					}
					break;

		case 'r':
		case 'R':
					switch(*(comnd+1))
					{
						case 'e':
						case 'E':
									switch (*(comnd+2))
									{
										case 'b':
										case 'B':
													return 3;
													break;
										default:	return -1;
									}
						break;
						default: return -1;
					}
					break;
		
		case 'h':
		case 'H':
					switch(*(comnd+1))
					{
						case 'l':
						case 'L':
									switch (*(comnd+2))
									{
										case 'p':
										case 'P':
													return 4;
													break;
										default:	return -1;
									}
						break;
						default: return -1;
					}
					break;
		case 'd': // About AVOS ....
		case 'D':
					switch(*(comnd+1))
					{
						case 'a':
						case 'A':
									switch (*(comnd+2))
									{
										case 't':
										case 'T':
													return 5; // Date 
													break;
										default:
													return -1;
									}
									break;
						default:
									return -1;
					}

					break;
		
		default: return -1;
	}

}

									

/*****************************************************************************
*****************************************************************************/
void init_keyboard(void)
{
	static unsigned char buffers[KBD_BUF_SIZE * MAX_VC];
/**/
	int i;

	for(i = 0; i < MAX_VC; i++)
	{
		_vc[i].keystrokes.data = buffers + KBD_BUF_SIZE * i;
		_vc[i].keystrokes.size = KBD_BUF_SIZE;
	}
	kprintf("Initializing keyboard: %u buffers, %u bytes each\n",
		MAX_VC, KBD_BUF_SIZE);
}
